from os import close, makedirs, walk
from pathlib import Path
from shutil import copy
from tempfile import mkdtemp, mkstemp
from typing import List

from pdistx.utils.path import fnmatch_any, rmpath
from pdistx.utils.zip import zipit

from .transform import variant_transform


def perform(
    source: Path,
    target: Path,
    definitions: dict,
    filters: List[Path],
    zip_: Path,
):
    # ensure pre-conditions
    assert source.is_file() or source.is_dir(), 'source path is expected to be a file or directory'

    if zip_:
        assert not target.is_absolute(), 'target path is expected to be relative'

    # list of temporary files and folders
    tmps: List[Path] = []

    # temporary paths get cleaned automatically at the end of this block
    try:

        # purging target or zip
        if zip_:
            print(f'Purging {zip_}...')
            rmpath(zip_)
        else:
            print(f'Purging {target}...')
            rmpath(target)

        # create target path
        if zip_:
            # create interim directory or file
            if source.is_dir():
                intermediate = Path(mkdtemp())
            else:
                tmp_fh, intermediate = mkstemp()
                close(tmp_fh)  # dump handle right away
                intermediate = Path(intermediate)

            # add to tmp paths for cleanup
            tmps.append(intermediate)
        else:
            intermediate = target

        # handle source folder
        print(f'Processing {source}...')

        if source.is_dir():

            # ensure target directory exists
            makedirs(intermediate, exist_ok=True)

            # process all files
            for source_folder, folders, files in walk(source, followlinks=True):

                # ensure target directory exists
                source_folder = Path(source_folder)
                target_folder = intermediate.joinpath(source_folder.relative_to(source))
                makedirs(target_folder, exist_ok=True)

                # filter entries to be ignored (folders need to be modified in-place to take effect for os.walk)
                def _folder_filter(folder: Path):
                    return not fnmatch_any(folder.name, ['__pycache__', '.git']) and folder not in filters

                def _file_filter(file: Path):
                    return not fnmatch_any(file.name, ['*.pyc']) and file not in filters

                folders[:] = [folder for folder in folders if _folder_filter(source_folder.joinpath(folder))]
                files = [file for file in files if _file_filter(source_folder.joinpath(file))]

                # transform or copy file
                for file in files:
                    source_file = source_folder.joinpath(file)
                    target_file = target_folder.joinpath(file)

                    if file.endswith('.py'):
                        variant_transform(source_file, target_file, definitions)
                    else:
                        copy(source_file, target_file, follow_symlinks=True)

        # handle source file
        else:

            # ensure parent target directory exists
            makedirs(intermediate.parent, exist_ok=True)

            # transform file
            variant_transform(source, intermediate, definitions)

        # zip intermediate path to zip path
        if zip_:
            zipit(intermediate, zip_, target)

    finally:
        # clean up temporary folders
        for path in tmps:
            print(f'Purging {path}...')
            rmpath(path)



import argparse
import sys
from pathlib import Path
from traceback import print_tb
from typing import List

from pvendor.process import perform


def main(argv: List[str] = sys.argv[1:]):

    parser = argparse.ArgumentParser(prog='pvendor')

    parser.add_argument(
        '-r',
        dest='requirements',
        metavar='requirements',
        action='append',
        default=[],
        help='install packages from requirements.txt',
    )

    parser.add_argument(
        '-s',
        dest='source',
        metavar='source',
        action='append',
        default=[],
        help='copy modules from source folder',
    )

    parser.add_argument(
        '-p',
        dest='pip',
        metavar='pip',
        default='pip',
        help='pip command (defaults to pip)',
    )

    parser.add_argument(
        '-k',
        dest='keep',
        metavar='keep',
        action='append',
        default=[],
        help='files or folders to be kept in the target folder (defaults to requirements.txt and .gitignore)',
    )

    parser.add_argument(
        '-z',
        dest='zip',
        metavar='zip',
        default=None,
        help='zip file path (target becomes relative path within zip file)',
    )

    parser.add_argument(
        'target',
        help='target folder (will be cleared, except for the ones to be kept)',
    )

    args = parser.parse_args(argv)

    try:
        perform(
            [Path(req) for req in args.requirements],
            args.pip,
            [Path(req) for req in args.source],
            Path(args.target),
            args.keep if args.keep else ['requirements.txt', '.gitignore'],
            Path(args.zip) if args.zip else None,
        )
    except Exception as ex:
        print(f'ERROR: {ex}')
        print_tb(ex.__traceback__)
        sys.exit(1)

    sys.exit(0)


if __name__ == '__main__':
    main()



import ast


class _HasAbsoluteImportOfModuleCheck(ast.NodeTransformer):

    def __init__(self, module):
        self._module = module
        self.has_absolute_import_of_module = False
        super().__init__()

    def visit_Import(self, node: ast.Import):
        for name in node.names:
            if name.name == self._module or name.name.startswith(self._module + '.'):
                self.has_absolute_import_of_module = True
        return node

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.level == 0:
            if node.module == self._module or node.module.startswith(self._module + '.'):
                self.has_absolute_import_of_module = True
        return node


def has_absolute_import_of_module(source, module):
    visitor = _HasAbsoluteImportOfModuleCheck(module)
    visitor.visit(ast.parse(source, type_comments=True))
    return visitor.has_absolute_import_of_module


class _HasRelativeImportCheck(ast.NodeTransformer):

    def __init__(self):
        self.has_relative_import = False
        super().__init__()

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.level != 0:
            self.has_relative_import = True
        return node


def has_relative_import(source):
    visitor = _HasRelativeImportCheck()
    visitor.visit(ast.parse(source, type_comments=True))
    return visitor.has_relative_import



from os import listdir, makedirs, walk
from pathlib import Path
from shutil import copy
from subprocess import check_call
from tempfile import mkdtemp
from typing import List

from pdistx.utils.path import fnmatch_any, rmpath
from pdistx.utils.zip import zipit

from .transform import import_transform


def perform(
    requirements: List[Path],
    pip: str,
    sources: List[Path],
    target: Path,
    keep: List[str],
    zip_: Path,
):
    # ensure pre-conditions
    for requirement in requirements:
        assert requirement.is_file(), 'requirements.txt is expected to be a file'

    for source in sources:
        assert source.is_dir(), 'source path is expected to be a directory'

    if zip_:
        assert not target.is_absolute(), 'target path is expected to be relative'

    # list of temporary files and folders
    tmps: List[Path] = []

    # temporary paths get cleaned automatically at the end of this block
    try:

        # detect requirements.txt in target folder
        if not zip_:
            requirement = target.joinpath('requirements.txt')
            if requirement.is_file():
                requirements.append(requirement)

        # create a source folder for each requirements
        for requirement in requirements:
            # create temp folder
            install_folder = Path(mkdtemp())
            tmps.append(install_folder)
            sources.append(install_folder)

            # install packages into temp folder
            print(f'Installing {requirement} to {install_folder}...')
            check_call([pip, 'install', '--upgrade', '--requirement', requirement, '--target', str(install_folder)])

        # clean target folder
        if zip_:
            print(f'Purging {zip_}...')
            rmpath(zip_)
        else:
            print(f'Purging {target}...')
            if target.is_dir():
                # if target is not a zip, remove all except the entries to be kept
                if not zip_:
                    for name in listdir(target):
                        if not fnmatch_any(name, keep):
                            path = target.joinpath(name)
                            rmpath(path)

                # if target is a zip, remove entire folder
                else:
                    rmpath(target)
            else:
                # remove an existing target file in any case
                rmpath(target)

        # build dictionary of modules
        # pylint: disable=unsubscriptable-object
        modules: dict[str, Path] = {}

        for source in sources:
            for entry in listdir(source):
                # detect module name
                name = None
                path = source.joinpath(entry)

                if path.is_dir():
                    if fnmatch_any(entry, ['*.dist-info', 'bin', '__pycache__', '.git']):
                        continue
                    name = entry

                else:
                    if not entry.endswith('.py'):
                        continue
                    name = entry[:-3]

                # add to module dictionary
                if name in modules:
                    print(f'Warning: multiple copies of {name} detected, skipping redundant one!')
                    continue

                # add to dictionary
                modules[name] = path

        # create target path
        if zip_:
            intermediate = Path(mkdtemp())
            tmps.append(intermediate)
        else:
            makedirs(target, exist_ok=True)
            intermediate = target

        # copy and transform all module files
        for name, source in modules.items():
            print(f'Processing {name} from {source}...')

            # handle directory case
            if source.is_dir():
                for source_folder, folders, files in walk(source, followlinks=True):
                    # filter entries to be ignored (folders need to be modified in-place to take effect for os.walk)
                    folders[:] = [folder for folder in folders if not fnmatch_any(folder, ['__pycache__', '.git'])]
                    files = [file for file in files if not fnmatch_any(file, ['*.pyc'])]

                    # ensure sub target directory exists
                    source_folder = Path(source_folder)
                    package_folder = source_folder.relative_to(source)
                    target_folder = intermediate.joinpath(name, package_folder)
                    makedirs(target_folder, exist_ok=True)

                    # transform or copy files
                    level = len(package_folder.parts) + 1

                    for file in files:
                        source_file = source_folder.joinpath(file)
                        target_file = target_folder.joinpath(file)

                        if file.endswith('.py'):
                            import_transform(source_file, target_file, level, list(modules.keys()))
                        else:
                            copy(source_file, target_file, follow_symlinks=True)

            # handle file case
            else:
                import_transform(source, intermediate.joinpath(name + '.py'), 1, list(modules.keys()))

        # create empty init file in target folder
        with open(intermediate.joinpath('__init__.py'), 'w'):
            pass

        # zip temporary target path to actual target path
        if zip_:
            zipit(intermediate, zip_, target)

    finally:
        # clean up temporary folders
        for path in tmps:
            print(f'Purging {path}...')
            rmpath(path)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pdistx",
    author="3D Ninjas GmbH",
    author_email="niklas@3dninjas.io",
    description="A toolset for distributing Python projects in a convenient way.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/3dninjas/pdistx",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    license="GPLv3",
    python_requires=">=3.9",
    install_requires=[],
    entry_points={
        "console_scripts": [
            "pdistx=pdistx.__main__:main",
            "pvendor=pvendor.__main__:main",
            "pvariant=pvariant.__main__:main",
            "ppack=ppack.__main__:main",
        ],
    },
)
